from pathlib import Path
from logging import Handler, Formatter, LogRecord
from collections.abc import Callable
from functools import partial

from ...app import PyRandyOSApp
from ...logging import (
    INFO, LOGSTDERR, LOGSTDOUT, LOGTQDM, WARNING, ERROR, CRITICAL, DEBUG,
    log_func_call, get_logger, get_loglevel_num_name, make_log_record,
    log_debuglow2, log_debug,
)
from ...utils.log import (
    check_loglevel, MillisecondFormatter, LogMultiFormatter,
)
from ...config.keys import BASE_LOG_PATH_KEY
from ..qt import (
    QStatusBar, QLabel, QMainWindow, Qt, QMouseEvent, QPoint, QMenu,
)
from ..callback import qt_callback
from ..gui_app import get_gui_app
from ..dialogs.log import LogDialog
from . import QtWidgetWrapper, GuiWindowLikeParentType

StatusMsgStyleDict = dict[str | int, str]
StatusMsgClearDict = dict[str | int, bool]

DEFAULT_STATUSMSG_STYLES: StatusMsgStyleDict = {
    DEBUG: "background-color: blue; color: white;",
    # INFO: "background-color: green; color: white;",
    WARNING: "background-color: yellow; color: black;",
    ERROR: "background-color: red; color: white;",
    CRITICAL: "background-color: magenta; color: black; font-weight: bold;",
}

DEFAULT_STATUSMSG_CLEAR: StatusMsgClearDict = {
    DEBUG: False,
    INFO: True,
    WARNING: True,
    ERROR: True,
    CRITICAL: True,
}

DEFAULT_STATUSMSG_FMT = LogMultiFormatter({
    INFO: MillisecondFormatter("%(message)s"),
    'DEFAULT': MillisecondFormatter("%(levelname)s: %(message)s"),
})


class StatusMsgLogFilter:
    def filter(self, record: LogRecord):
        return record.levelno not in (LOGSTDERR, LOGSTDOUT, LOGTQDM)


class StatusMsgLogHandler(Handler):
    def __init__(self, lbl: QLabel, level: int = 0,
                 style_dict: StatusMsgStyleDict = DEFAULT_STATUSMSG_STYLES,
                 clear_prev_message_callback: callable = None,
                 clear_dict: StatusMsgClearDict = DEFAULT_STATUSMSG_CLEAR,
                 reset_style_callback: Callable = None,
                 max_lines: int = 3):
        super().__init__(level)
        self.lbl = lbl
        self.reset_style_callback = reset_style_callback
        self.style_dict = style_dict
        self.clear_prev_message_callback = clear_prev_message_callback
        self.clear_dict = clear_dict
        self.max_lines = max_lines
        self.last_msg = None

    # def flush(self):
    #     """
    #     Flushes the stream.
    #     """
    #     self.acquire()
    #     try:
    #         if self.stream and hasattr(self.stream, "flush"):
    #             self.stream.flush()
    #     finally:
    #         self.release()

    def update_label(self, msg_or_record: str | LogRecord = None,
                     level: int | str = INFO, clear: bool = False,
                     get_last: bool = False):
        last_msg = self.last_msg
        get_last_only = get_last and not msg_or_record and not clear
        if get_last_only:
            return last_msg

        if msg_or_record:
            record = (msg_or_record if isinstance(msg_or_record, LogRecord)
                      else make_log_record(level, msg_or_record))
            msg = self.format(record)
        else:
            clear = clear or not get_last
            record = None
            msg = ""

        clearcb = self.clear_prev_message_callback
        if clearcb and clear or (record and self.check_level_clear(record)):
            clearcb()

        self.update_status_style(record)
        if msg:
            lines = msg.splitlines()
            self.lbl.setText('\n'.join(lines[:self.max_lines]))
        else:
            self.lbl.clear()

        get_gui_app().process_events()
        self.last_msg = msg
        if get_last:
            return last_msg

    def emit(self, record: LogRecord):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            self.update_label(record)
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

    def check_level_clear(self, record: LogRecord):
        return (self.clear_dict.get(record.levelno, False) if self.clear_dict
                else False)

    def update_status_style(self, record: LogRecord = None):
        style = self.style_dict.get(record.levelno, None) if record else None
        lbl = self.lbl
        ss = f"QLabel {{ {style} }}" if style else ""
        cb = self.reset_style_callback
        if style or not cb:
            lbl.setStyleSheet(ss)
        else:
            cb()


class LoggingStatusBarWidget(QtWidgetWrapper[QStatusBar]):
    @log_func_call
    def __init__(self, parent: GuiWindowLikeParentType):  # , *qtobj_args):
        self._log_dialog: LogDialog = None
        super().__init__(parent)  # , *qtobj_args)

    @log_func_call
    def create_qtobj(self):  # , *qtobj_args):
        qtwin: QMainWindow = self.gui_parent.qtobj

        statusbar = QStatusBar(qtwin)
        qtwin.setStatusBar(statusbar)
        self.status_bar = statusbar

        status_msg_lbl = QLabel(statusbar)
        status_msg_lbl.mousePressEvent = self.click_status_msg
        statusbar.addWidget(status_msg_lbl, 1)
        self.status_msg_lbl = status_msg_lbl

        # Set up context menu for right-click
        status_msg_lbl.setContextMenuPolicy(Qt.CustomContextMenu)
        status_context_signal = status_msg_lbl.customContextMenuRequested
        status_context_signal.connect(
            qt_callback(self.show_status_context_menu))

        level = PyRandyOSApp.get('statusbar_log_level', INFO)
        msglog = self.add_statusmsg_log_handler(level, statusbar.clearMessage)
        updater = msglog.update_label
        self.status_msg_updater = updater
        updater('Ready')

        return statusbar

    @log_func_call
    def add_statusmsg_log_handler(
        self,
        level: int | str = INFO,
        clear_prev_message_callback: Callable = None,
        fmt: Formatter = DEFAULT_STATUSMSG_FMT,
        style_dict: StatusMsgStyleDict = DEFAULT_STATUSMSG_STYLES,
        reset_style_callback: Callable = None,
        clear_dict: StatusMsgClearDict = DEFAULT_STATUSMSG_CLEAR,
    ):
        lbl = self.status_msg_lbl
        check_loglevel(level)
        lnum, _ = get_loglevel_num_name(level)
        msglog = StatusMsgLogHandler(lbl, lnum, style_dict,
                                     clear_prev_message_callback, clear_dict,
                                     reset_style_callback)
        msglog.setFormatter(fmt)
        msglog.addFilter(StatusMsgLogFilter())
        get_logger().root.addHandler(msglog)
        return msglog

    @log_func_call
    def update_status_bar_msg(self, msg: str, level: int | str = INFO,
                              temp: bool = False, timeout_ms: int = 0):
        if temp:
            bar = self.status_bar
            bar.showMessage(msg, timeout_ms)
        else:
            self.status_msg_updater(msg, level)

    @log_func_call
    def clear_status_message(self):
        """Clear the status bar message."""
        self.status_bar.clearMessage()
        self.status_msg_updater()
        log_debuglow2("Status message cleared")

    @log_func_call
    def copy_status_message(self, msg: str):
        """Copy the current status message to clipboard."""
        if msg:
            get_gui_app().qtobj.clipboard().setText(msg)
            log_debug(f"Status message copied: {msg}")

    @log_func_call
    def click_status_msg(self, event: QMouseEvent = None):
        if not event or event.button() == Qt.LeftButton:
            self.show_log_dialog()

    @log_func_call
    def get_log_path(self):
        logpath = PyRandyOSApp.get(BASE_LOG_PATH_KEY, None)
        return Path(logpath) if logpath else None

    @log_func_call
    def show_log_dialog(self, log_path: Path = None):
        log_path = log_path or self.get_log_path()
        if log_path:
            # Close existing dialog if open
            dialog = self._log_dialog
            if dialog:
                try:
                    dialog.gui_view.qtobj.close()
                except RuntimeError:
                    # Dialog was already closed/destroyed
                    pass

            # Create and show new dialog
            dialog = LogDialog(self, log_path)
            self._log_dialog = dialog
            dialog.show()

    @log_func_call
    def get_current_status_msg(self):
        return (self.status_bar.currentMessage()
                or self.status_msg_updater(get_last=True))

    @log_func_call
    def show_status_context_menu(self, position: QPoint):
        """Show context menu for status bar message label."""
        pres = self.gui_pres
        msg = pres.get_current_status_msg()

        menu = QMenu()
        if msg:
            clear = menu.addAction("Clear Message")
            clear.triggered.connect(qt_callback(pres.clear_status_message))
            copy = menu.addAction("Copy Message")
            copy.triggered.connect(
                qt_callback(partial(pres.copy_status_message, msg)))
            menu.addSeparator()

        logpath = pres.get_log_path()
        logexists = logpath and logpath.exists()
        status_dialog = menu.addAction("Show Log History Dialog")
        status_dialog.triggered.connect(
            qt_callback(partial(pres.show_log_dialog, logpath)))
        status_dialog.setEnabled(logexists)

        # Show menu at the requested position
        menu.exec_(self.status_msg_lbl.mapToGlobal(position))



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
address_book = 62137
address_card = 62139
bell = 61683
bell_slash = 61942
bookmark = 61486
building = 61869
calendar = 61747
calendar_check = 62068
calendar_days = 61555
calendar_minus = 62066
calendar_plus = 62065
calendar_xmark = 62067
chart_bar = 61568
chess_bishop = 62522
chess_king = 62527
chess_knight = 62529
chess_pawn = 62531
chess_queen = 62533
chess_rook = 62535
circle = 61713
circle_check = 61528
circle_dot = 61842
circle_down = 62296
circle_left = 62297
circle_pause = 62091
circle_play = 61764
circle_question = 61529
circle_right = 62298
circle_stop = 62093
circle_up = 62299
circle_user = 62141
circle_xmark = 61527
clipboard = 62248
clock = 61463
clone = 62029
closed_captioning = 61962
comment = 61557
comment_dots = 62637
comments = 61574
compass = 61774
copy = 61637
copyright = 61945
credit_card = 61597
envelope = 61664
envelope_open = 62134
eye = 61550
eye_slash = 61552
face_angry = 62806
face_dizzy = 62823
face_flushed = 62841
face_frown = 61721
face_frown_open = 62842
face_grimace = 62847
face_grin = 62848
face_grin_beam = 62850
face_grin_beam_sweat = 62851
face_grin_hearts = 62852
face_grin_squint = 62853
face_grin_squint_tears = 62854
face_grin_stars = 62855
face_grin_tears = 62856
face_grin_tongue = 62857
face_grin_tongue_squint = 62858
face_grin_tongue_wink = 62859
face_grin_wide = 62849
face_grin_wink = 62860
face_kiss = 62870
face_kiss_beam = 62871
face_kiss_wink_heart = 62872
face_laugh = 62873
face_laugh_beam = 62874
face_laugh_squint = 62875
face_laugh_wink = 62876
face_meh = 61722
face_meh_blank = 62884
face_rolling_eyes = 62885
face_sad_cry = 62899
face_sad_tear = 62900
face_smile = 61720
face_smile_beam = 62904
face_smile_wink = 62682
face_surprise = 62914
face_tired = 62920
file = 61787
file_audio = 61895
file_code = 61897
file_excel = 61891
file_image = 61893
file_lines = 61788
file_pdf = 61889
file_powerpoint = 61892
file_video = 61896
file_word = 61890
file_zipper = 61894
flag = 61476
floppy_disk = 61639
folder = 61563
folder_closed = 57733
folder_open = 61564
font_awesome = 62132
futbol = 61923
gem = 62373
hand = 62038
hand_back_fist = 62037
hand_lizard = 62040
hand_peace = 62043
hand_point_down = 61607
hand_point_left = 61605
hand_point_right = 61604
hand_point_up = 61606
hand_pointer = 62042
hand_scissors = 62039
hand_spock = 62041
handshake = 62133
hard_drive = 61600
heart = 61444
hospital = 61688
hourglass = 62036
hourglass_half = 62034
id_badge = 62145
id_card = 62146
image = 61502
images = 62210
keyboard = 61724
lemon = 61588
life_ring = 61901
lightbulb = 61675
map = 62073
message = 62074
money_bill_1 = 62417
moon = 61830
newspaper = 61930
note_sticky = 62025
object_group = 62023
object_ungroup = 62024
paper_plane = 61912
paste = 61674
pen_to_square = 61508
rectangle_list = 61474
rectangle_xmark = 62480
registered = 62045
share_from_square = 61773
snowflake = 62172
square = 61640
square_caret_down = 61776
square_caret_left = 61841
square_caret_right = 61778
square_caret_up = 61777
square_check = 61770
square_full = 62556
square_minus = 61766
square_plus = 61694
star = 61445
star_half = 61577
star_half_stroke = 62912
sun = 61829
thumbs_down = 61797
thumbs_up = 61796
trash_can = 62189
user = 61447
window_maximize = 62160
window_minimize = 62161
window_restore = 62162



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path

from ....logging import log_func_call
from ...qt import QPalette, QColor
if TYPE_CHECKING:
    from ...gui_app import QtApp

from ....utils.encoding import read_text_utf8
# from ...icons.iconfont.icon import IconSpec
# from ....utils.json import load_jsonc
# from ...utils import qicon_to_data_uri
# from ..qss import merge_qss_properties

ASSETS_DIR = Path(__file__).parent.parent / "assets"


@log_func_call
def vibedark(app: QtApp):
    palette = QPalette()
    palette.setColor(QPalette.Window, QColor(53, 53, 53))
    palette.setColor(QPalette.WindowText, QColor(255, 255, 255))
    palette.setColor(QPalette.Base, QColor(35, 35, 35))
    palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ToolTipBase, QColor(255, 255, 255))
    palette.setColor(QPalette.ToolTipText, QColor(255, 255, 255))
    palette.setColor(QPalette.Text, QColor(255, 255, 255))
    palette.setColor(QPalette.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ButtonText, QColor(255, 255, 255))
    palette.setColor(QPalette.BrightText, QColor(255, 0, 0))
    palette.setColor(QPalette.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.HighlightedText, QColor(0, 0, 0))
    # Supplement missing roles from dark()
    palette.setColor(QPalette.Light, QColor(180, 180, 180))
    palette.setColor(QPalette.Midlight, QColor(90, 90, 90))
    palette.setColor(QPalette.Dark, QColor(35, 35, 35))
    palette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    palette.setColor(QPalette.LinkVisited, QColor(80, 80, 80))
    # Disabled state roles
    palette.setColor(QPalette.Disabled, QPalette.WindowText,
                     QColor(127, 127, 127))
    palette.setColor(QPalette.Disabled, QPalette.Text, QColor(127, 127, 127))
    palette.setColor(QPalette.Disabled, QPalette.ButtonText,
                     QColor(127, 127, 127))
    palette.setColor(QPalette.Disabled, QPalette.Highlight, QColor(80, 80, 80))
    palette.setColor(QPalette.Disabled, QPalette.HighlightedText,
                     QColor(127, 127, 127))

    # qss_file = ASSETS_DIR/"vibedark.qss"
    qss_file = ASSETS_DIR/"vibedark.qss"
    qss = read_text_utf8(qss_file)
    # qss = merge_qss_icons(qss)

    # qss_dump = PyRandyOSApp.mkdir_temp()/'vibedark-full.qss'
    # qss_dump.write_text(qss, encoding='utf-8')

    app.style().unpolish(app)
    app.setStyle("Fusion")
    app.setPalette(palette)
    app.setStyleSheet(qss)



"""
This module is intended for internal use only.  No user documentation is
provided at this time.  Use at your own discretion.
"""
from ctypes import c_uint

try:  # pragma: no cover
    from ctypes import WinDLL, WinError, get_last_error, OleDLL  # noqa: F401
    from ctypes.wintypes import (
        DWORD, HANDLE, LPDWORD, LPVOID,
        # LPCWSTR, ULONG, PHANDLE, PULONG,
    )
    from win32wnet import WNetOpenEnum, WNetCloseEnum, WNetEnumResource  # noqa: F401, E501
    from win32netcon import (  # noqa: F401
        RESOURCETYPE_DISK, RESOURCE_CONNECTED, RESOURCEDISPLAYTYPE_NETWORK,
        RESOURCEDISPLAYTYPE_DOMAIN, RESOURCEDISPLAYTYPE_SERVER,
        RESOURCEDISPLAYTYPE_SHARE, RESOURCEDISPLAYTYPE_DIRECTORY,
        RESOURCEDISPLAYTYPE_GENERIC,
    )
except ImportError:  # pragma: no cover
    IS_WIN_CTYPES = False
else:  # pragma: no cover
    IS_WIN_CTYPES = True

    kernel32 = WinDLL("kernel32", use_last_error=True)
    WriteConsoleW = kernel32.WriteConsoleW
    WriteConsoleW.argtypes = (HANDLE, LPVOID, DWORD, LPDWORD, LPVOID)
    GetConsoleMode = kernel32.GetConsoleMode
    GetConsoleMode.argtypes = (HANDLE, LPDWORD)

    shcore = OleDLL('shcore', use_last_error=True)
    SetProcessDpiAwareness = shcore.SetProcessDpiAwareness
    SetProcessDpiAwareness.argtypes = (c_uint,)

    RESOURCEDISPLAYTYPES = {
        # The resource is a network provider.
        RESOURCEDISPLAYTYPE_NETWORK: 'RESOURCEDISPLAYTYPE_NETWORK',
        # The resource is a collection of servers.
        RESOURCEDISPLAYTYPE_DOMAIN: 'RESOURCEDISPLAYTYPE_DOMAIN',
        # The resource is a server.
        RESOURCEDISPLAYTYPE_SERVER: 'RESOURCEDISPLAYTYPE_SERVER',
        # The resource is a share point.
        RESOURCEDISPLAYTYPE_SHARE: 'RESOURCEDISPLAYTYPE_SHARE',
        # The resource is a directory.
        RESOURCEDISPLAYTYPE_DIRECTORY: 'RESOURCEDISPLAYTYPE_DIRECTORY',
        # The resource type is unspecified.
        # This value is used by network providers that do not specify
        # resource types.
        RESOURCEDISPLAYTYPE_GENERIC: 'RESOURCEDISPLAYTYPE_GENERIC',
    }



# This file is auto-generated by pyrandyos/gui/icons/update_from_spec.py  # noqa: E501
# Do not edit.
ic_fluent_accessibility_checkmark_32_light = 57344
ic_fluent_add_32_light = 57345
ic_fluent_alert_32_light = 57346
ic_fluent_app_folder_32_light = 57347
ic_fluent_app_generic_32_light = 57348
ic_fluent_archive_32_light = 57349
ic_fluent_archive_settings_32_light = 57350
ic_fluent_arrow_clockwise_32_light = 57351
ic_fluent_arrow_down_32_light = 57352
ic_fluent_arrow_download_32_light = 57353
ic_fluent_arrow_forward_32_light = 57354
ic_fluent_arrow_hook_down_left_32_light = 57355
ic_fluent_arrow_hook_down_right_32_light = 57356
ic_fluent_arrow_hook_up_left_32_light = 57357
ic_fluent_arrow_hook_up_right_32_light = 57358
ic_fluent_arrow_redo_32_light = 57359
ic_fluent_arrow_reply_32_light = 57360
ic_fluent_arrow_reply_all_32_light = 57361
ic_fluent_arrow_undo_32_light = 57362
ic_fluent_attach_32_light = 57363
ic_fluent_auto_fit_32_light = 57364
ic_fluent_auto_fit_width_32_light = 57365
ic_fluent_autocorrect_32_light = 57366
ic_fluent_breakout_room_32_light = 57367
ic_fluent_broom_32_light = 57368
ic_fluent_calendar_3_day_32_light = 57369
ic_fluent_calendar_clock_32_light = 57370
ic_fluent_calendar_data_bar_32_light = 57371
ic_fluent_calendar_day_32_light = 57372
ic_fluent_calendar_edit_32_light = 57373
ic_fluent_calendar_empty_32_light = 57374
ic_fluent_calendar_ltr_32_light = 57375
ic_fluent_calendar_month_32_light = 57376
ic_fluent_calendar_multiple_32_light = 57377
ic_fluent_calendar_pattern_32_light = 57378
ic_fluent_calendar_reply_32_light = 57379
ic_fluent_calendar_sparkle_32_light = 57380
ic_fluent_calendar_todo_32_light = 57381
ic_fluent_calendar_work_week_32_light = 57382
ic_fluent_chat_32_light = 57383
ic_fluent_checkmark_32_light = 57384
ic_fluent_checkmark_circle_32_light = 57385
ic_fluent_classification_32_light = 57386
ic_fluent_clipboard_paste_32_light = 57387
ic_fluent_clock_32_light = 57388
ic_fluent_clock_alarm_32_light = 57389
ic_fluent_color_32_light = 57390
ic_fluent_color_fill_32_light = 57391
ic_fluent_color_fill_accent_32_light = 57392
ic_fluent_comment_32_light = 57393
ic_fluent_comment_add_32_light = 57394
ic_fluent_compose_32_light = 57395
ic_fluent_copy_32_light = 57396
ic_fluent_crop_32_light = 57397
ic_fluent_cursor_32_light = 57398
ic_fluent_cut_32_light = 57399
ic_fluent_delete_32_light = 57400
ic_fluent_dismiss_32_light = 57401
ic_fluent_dismiss_circle_32_light = 57402
ic_fluent_document_24_light = 57403
ic_fluent_document_28_light = 57404
ic_fluent_document_32_light = 57405
ic_fluent_document_48_light = 57406
ic_fluent_document_lightning_32_light = 57407
ic_fluent_document_signature_32_light = 57408
ic_fluent_document_sparkle_24_light = 57409
ic_fluent_document_sparkle_28_light = 57410
ic_fluent_document_sparkle_32_light = 57411
ic_fluent_document_sparkle_48_light = 57412
ic_fluent_door_arrow_right_32_light = 57413
ic_fluent_edit_32_light = 57414
ic_fluent_emoji_32_light = 57415
ic_fluent_eye_32_light = 57416
ic_fluent_eye_off_32_light = 57417
ic_fluent_filter_32_light = 57418
ic_fluent_flag_32_light = 57419
ic_fluent_flag_off_32_light = 57420
ic_fluent_flash_32_light = 57421
ic_fluent_folder_arrow_right_32_light = 57422
ic_fluent_folder_mail_32_light = 57423
ic_fluent_hand_draw_32_light = 57424
ic_fluent_history_32_light = 57425
ic_fluent_image_add_32_light = 57426
ic_fluent_image_alt_text_32_light = 57427
ic_fluent_image_copy_32_light = 57428
ic_fluent_image_reflection_32_light = 57429
ic_fluent_image_shadow_32_light = 57430
ic_fluent_immersive_reader_32_light = 57431
ic_fluent_important_32_light = 57432
ic_fluent_lasso_32_light = 57433
ic_fluent_layout_column_two_32_light = 57434
ic_fluent_layout_column_two_focus_left_32_light = 57435
ic_fluent_layout_column_two_focus_right_32_light = 57436
ic_fluent_layout_row_two_32_light = 57437
ic_fluent_layout_row_two_focus_top_32_light = 57438
ic_fluent_layout_row_two_focus_top_settings_32_light = 57439
ic_fluent_layout_row_two_settings_32_light = 57440
ic_fluent_lightbulb_32_light = 57441
ic_fluent_link_32_light = 57442
ic_fluent_lock_closed_32_light = 57443
ic_fluent_lock_open_32_light = 57444
ic_fluent_mail_32_light = 57445
ic_fluent_mail_alert_32_light = 57446
ic_fluent_mail_arrow_clockwise_32_light = 57447
ic_fluent_mail_arrow_double_back_32_light = 57448
ic_fluent_mail_copy_32_light = 57449
ic_fluent_mail_edit_32_light = 57450
ic_fluent_mail_list_32_light = 57451
ic_fluent_mail_multiple_32_light = 57452
ic_fluent_mail_read_32_light = 57453
ic_fluent_mail_read_multiple_32_light = 57454
ic_fluent_mail_rewind_32_light = 57455
ic_fluent_mail_settings_32_light = 57456
ic_fluent_mail_template_32_light = 57457
ic_fluent_mail_unread_32_light = 57458
ic_fluent_mic_32_light = 57459
ic_fluent_molecule_32_light = 57460
ic_fluent_note_32_light = 57461
ic_fluent_options_32_light = 57462
ic_fluent_paint_brush_32_light = 57463
ic_fluent_panel_left_default_32_light = 57464
ic_fluent_panel_left_focus_right_32_light = 57465
ic_fluent_pen_sparkle_32_light = 57466
ic_fluent_people_32_light = 57467
ic_fluent_people_add_32_light = 57468
ic_fluent_people_community_32_light = 57469
ic_fluent_people_edit_32_light = 57470
ic_fluent_people_list_32_light = 57471
ic_fluent_people_settings_32_light = 57472
ic_fluent_people_sync_32_light = 57473
ic_fluent_person_32_light = 57474
ic_fluent_person_add_32_light = 57475
ic_fluent_person_available_32_light = 57476
ic_fluent_person_feedback_32_light = 57477
ic_fluent_person_mail_32_light = 57478
ic_fluent_person_prohibited_32_light = 57479
ic_fluent_person_suport_32_light = 57480
ic_fluent_phone_32_light = 57481
ic_fluent_picture_in_picture_32_light = 57482
ic_fluent_pin_32_light = 57483
ic_fluent_pin_off_32_light = 57484
ic_fluent_poll_32_light = 57485
ic_fluent_print_32_light = 57486
ic_fluent_question_32_light = 57487
ic_fluent_read_aloud_32_light = 57488
ic_fluent_rectangle_landscape_32_light = 57489
ic_fluent_rotate_left_32_light = 57490
ic_fluent_save_32_light = 57491
ic_fluent_send_clock_32_light = 57492
ic_fluent_settings_32_light = 57493
ic_fluent_share_32_light = 57494
ic_fluent_shield_error_32_light = 57495
ic_fluent_signature_32_light = 57496
ic_fluent_speaker_mute_32_light = 57497
ic_fluent_square_arrow_forward_32_light = 57498
ic_fluent_stamp_32_light = 57499
ic_fluent_star_add_32_light = 57500
ic_fluent_star_arrow_right_32_light = 57501
ic_fluent_sticker_32_light = 57502
ic_fluent_tab_add_32_light = 57503
ic_fluent_table_32_light = 57504
ic_fluent_table_alt_text_32_light = 57505
ic_fluent_table_cells_merge_32_light = 57506
ic_fluent_table_cells_split_32_light = 57507
ic_fluent_table_dismiss_32_light = 57508
ic_fluent_table_move_above_32_light = 57509
ic_fluent_table_move_below_32_light = 57510
ic_fluent_table_move_left_32_light = 57511
ic_fluent_table_move_right_32_light = 57512
ic_fluent_table_settings_32_light = 57513
ic_fluent_table_simple_32_light = 57514
ic_fluent_tag_32_light = 57515
ic_fluent_text_32_light = 57516
ic_fluent_text_clear_formatting_32_light = 57517
ic_fluent_text_collapse_32_light = 57518
ic_fluent_text_density_32_light = 57519
ic_fluent_text_edit_style_32_light = 57520
ic_fluent_text_expand_32_light = 57521
ic_fluent_textbox_align_top_left_32_light = 57522
ic_fluent_toolbox_32_light = 57523
ic_fluent_translate_32_light = 57524
ic_fluent_video_32_light = 57525
ic_fluent_video_clip_32_light = 57526
ic_fluent_weather_moon_32_light = 57527
ic_fluent_weather_sunny_32_light = 57528
ic_fluent_window_32_light = 57529
ic_fluent_wrench_screwdriver_32_light = 57530
ic_fluent_zoom_in_32_light = 57531
ic_fluent_zoom_out_32_light = 57532



# Nomenclature, Model-View-Presenter, and Qt concepts/hierarchy
# --------------------------------------------------------------
# * Qt terminology:
#   * Qt widget: any components of a GUI a user can see or interact with
#     * Qt window: a top-level widget with no parent other than the application
#       * Qt dialog: a transient window that is intended to be dismissed
#       * Qt *main* window: a window intended to persist and not be dismissed,
#         also dedicated structure for menus, toolbars, status bars, and docks
#   * Qt object: any object in the Qt framework, including non-widgets.
#     * note that PySide classes cannot add new attributes or they will crash
#       the Python interpreter as soon as you try to add a new attribute not
#       part of the base class.  PySide classes should only be subclassed to
#       override methods, not to add new attributes.  If you need to store
#       additional data outside of stack variables, use a wrapper class around
#       the Qt object
#
# * Model: data interface and application state.  These are concepts in the
#   general application library that are GUI-agnostic.
#
# * View: code to construct a visible GUI component (widget OR window).
#   * Qt widget/window wrappers are generally "views".
#   * generally only stores Qt objects and data about the structure/layout of
#     the GUI.  Other data should be stored in the Model.
#   * owned by a Presenter and considered internal to it.
# * Presenter: contains callbacks for user interaction with the View.
#   * owns and creates a View.
#   * this is the object that should be passed around and interacted with
#     by the models or other parts of the application.
#   * It need not have "presenter" in the name, and GUI components lacking any
#     qualifier should be assumed to be a Presenter.
#
# * PyRandyOS-specific nomenclature:
#   * PyRandyOS implements a hierarchical MVP structure.
#
#   * GUI Component: an abstract concept that encapsulates a View and
#     its Presenter.
#
#   * A "window", with no other qualifiers in the name, is assumed to correlate
#     to a Qt main window, i.e. not a dialog or orphaned widget.  This differs
#     from the Qt nomenclature, where a "window" is any top-level widget.
#     This is merely a convention to shorten names and avoid confusion.
#   * Top View: a window, dialog, or top-level widget.
#   * Top Pres: the Presenter of a Top View.
#
#   * GUI App: analogous to a Presenter, this is the object that wraps a
#     Qt QApplication and is the main entry point for the GUI.
#     * Is a GUI Container for Top Views spawned by the main entry point.

from typing import TypeVar, TYPE_CHECKING, Generic

from ..logging import log_func_call
from .qt import QWidget, QApplication

if TYPE_CHECKING:
    from .gui_app import QtApp, GuiApp
    from .widgets import GuiWidgetParentType


QtObject = TypeVar('QtObject', bound=QWidget | QApplication)
QtWidgetType = TypeVar('QtWidgetType', bound=QWidget)
GuiViewType = TypeVar('GuiViewType', bound='GuiView')
GuiPresType = TypeVar('GuiPresType', bound='GuiPresenter')


class GuiQtWrapper(Generic[QtObject]):
    @log_func_call
    def __init__(self, qtobj: QtObject):
        self._qtobj = qtobj

    @property
    def qtobj(self) -> QtObject:
        return self._qtobj


class CreateQtObjMixin(GuiQtWrapper[QtWidgetType], Generic[QtWidgetType]):
    @log_func_call
    def __init__(self, *qtobj_args, **qtobj_kwargs):
        super().__init__(self.create_qtobj(*qtobj_args, **qtobj_kwargs))

    @log_func_call
    def create_qtobj(self, *args, **kwargs) -> QtWidgetType:
        raise NotImplementedError('Abstract method not implemented')


class GuiComponent:
    @property
    def gui_app(self) -> 'GuiApp':
        from .gui_app import get_gui_app
        return get_gui_app()

    @property
    def qt_app(self) -> 'QtApp':
        app = self.gui_app
        if app:
            return app.qtobj

    @property
    def gui_view(self) -> 'GuiView':
        return getattr(self, '_gui_view', None)

    @property
    def gui_pres(self) -> 'GuiPresenter':
        return getattr(self, '_gui_pres', None)


class GuiPresenter(GuiComponent, Generic[GuiViewType]):
    @log_func_call
    def __init__(self, gui_parent: 'GuiWidgetParentType | None' = None,
                 *view_args, **view_kwargs):
        GuiComponent.__init__(self)
        self._gui_parent = gui_parent
        self._gui_view = self.create_gui_view(*view_args, **view_kwargs)

    @log_func_call
    def create_gui_view(self, *args, **kwargs) -> GuiViewType:
        raise NotImplementedError('Abstract method not implemented')

    @property
    def gui_parent(self):
        return self._gui_parent

    @property
    def gui_pres(self):
        return self

    @property
    def gui_view(self) -> GuiViewType:
        return self._gui_view


class GuiView(GuiComponent, Generic[GuiPresType]):
    @log_func_call
    def __init__(self, presenter: GuiPresType = None):
        GuiComponent.__init__(self)
        self._gui_pres = presenter

    @property
    def gui_pres(self) -> GuiPresType:
        return self._gui_pres

    @property
    def gui_view(self):
        return self


class GuiQtView(GuiView[GuiPresType], CreateQtObjMixin[QtWidgetType],
                Generic[GuiPresType, QtWidgetType]):
    @log_func_call
    def __init__(self, presenter: GuiPresType = None, *qtobj_args,
                 **qtobj_kwargs):
        GuiView.__init__(self, presenter)
        CreateQtObjMixin.__init__(self, *qtobj_args, **qtobj_kwargs)



from pathlib import Path

from pyrandyos.utils.stack import top_package_dir_path
from pyrandyos.gui.icons.iconfont.sources import THIRDPARTY_FONTSPEC
from pyrandyos.gui.icons.utils import legalize_iconname
from pyrandyos.gui.icons.iconfont.fontspec import (
    ICON_ASSETS_DIR, THIRDPARTY_DIR, IconFontSpec
)

SCRIPTPATH = Path(__file__)
HERE = SCRIPTPATH.parent
NAMES_PY = 'names.py'
PYRANDYOS_DIR = top_package_dir_path('pyrandyos')
PYRANDYOS_ROOT = PYRANDYOS_DIR.parent
SCRIPT_RELPATH = SCRIPTPATH.relative_to(PYRANDYOS_ROOT)
AUTOHEADER = f'''
# This file is auto-generated by {SCRIPT_RELPATH.as_posix()}  # noqa: E501
# Do not edit.
'''.lstrip()
INIT_TEMPLATE_FILE = HERE/'init_template.py'
INIT_TEMPLATE = INIT_TEMPLATE_FILE.read_text()


def update_from_spec():
    ICON_ASSETS_DIR.mkdir(exist_ok=True)

    for fontmod, fontspec in THIRDPARTY_FONTSPEC.items():
        fontspec.initialize(fontmod, ICON_ASSETS_DIR)
        relmodname = fontspec.relative_module_qualname
        modparts = relmodname.split('.')

        levels = 1
        p = THIRDPARTY_DIR
        for part in modparts:
            levels += 1
            p /= part
            p.mkdir(exist_ok=True)
            init = p/'__init__.py'
            init.touch()

        generate_init(p, levels, fontspec)
        generate_names(p, fontspec)


def generate_init(p: Path, levels: int, fontspec: IconFontSpec):
    init = p/'__init__.py'
    s = f'{AUTOHEADER}{INIT_TEMPLATE}'
    s = s.replace('FONTCLASSNAME', fontspec.classname)
    s = s.replace("'FONTSPECNAME'",
                  repr(fontspec.target_relative_class_qualname))
    s = s.replace('\nfrom .', f"\nfrom {'.'*levels}.")
    init.write_text(s)


def generate_names(p: Path, fontspec: IconFontSpec):
    outf = p/NAMES_PY
    s = str(AUTOHEADER)
    for k, v in fontspec.charmap.items():
        newname = legalize_iconname(k)
        line = f"{newname} = {v}"
        if len(line) > 79:
            line += "  # noqa: E501"

        s += f"{line}\n"

    outf.write_text(s)


if __name__ == "__main__":
    update_from_spec()



import sys
from pathlib import Path, PurePosixPath, PureWindowsPath, PurePath, WindowsPath

from ..logging import log_func_call, DEBUGLOW2
from .expandvars import expandvars, is_key_resolved
from .constants import IS_WIN32


@log_func_call(DEBUGLOW2, trace_only=True)
def get_equiv_pureposixpath(x: str | Path):
    """
    Returns a `PurePosixPath` object equivalent to the given input path.

    The Python standard library `pathlib.Path` class automatically assumes the
    paths it is passed are valid paths for the platform it is running on and
    does not try to infer anything from its own contents.  This function exists
    to provide a means to convert Windows paths to POSIX-style paths in a
    consistent way so that `Path` objects created from the results of this
    function operate correctly in a platform-independent way.

    No expansions or path resolution is performed by this function; the path is
    converted as is.

    Args:
        x (str  |  Path): input path

    Returns:
        PurePosixPath: path to `x` as a `PurePosixPath`, or None if `x` is None
            or empty
    """
    # this voodoo was found experimentally and forces all flavors of Windows
    # paths and slashes to unify and become POSIX-ified.  All these steps are
    # necessary for reasons that I can't recall, but I assure you that it
    # needs to stay like this.  Don't be tempted to simplify.
    if x:
        return PurePosixPath(PureWindowsPath(str(x)).as_posix())


@log_func_call(DEBUGLOW2, trace_only=True)
def pureposixpath_to_pathobj(ppp: PurePosixPath):
    """
    Converts a `PurePosixPath` object to a `Path` object, expanding user
    directories if necessary.

    Args:
        ppp (PurePosixPath): path to convert to `Path`

    Returns:
        Path: path pointed to by `ppp` with user directory expanded if needed
    """
    # the as_posix() here is required to workaround a "feature" where the slash
    # after the drive letter in windows gets dropped otherwise
    if ppp:
        pathobj = Path(ppp.as_posix()).expanduser()
        if (isinstance(pathobj, WindowsPath) and pathobj.drive
                and not pathobj.root):
            pathobj = Path(f'{pathobj}/')

        return pathobj


@log_func_call(DEBUGLOW2, trace_only=True)
def pureposixpath_to_resolved_pathobj(ppp: PurePosixPath):
    """
    Converts a `PurePosixPath` object to an absolute `Path` object, expanding
    user directories if necessary.

    Args:
        ppp (PurePosixPath): path to convert to an absolute `Path`

    Returns:
        Path: absolute path pointed to by `ppp` with user directory expanded
            if needed
    """
    pathobj = pureposixpath_to_pathobj(ppp)
    # if we have a leading unexpanded variable at this point, we can't rule out
    # that the variable might have an absolute path in it, so don't try to
    # resolve it any further.
    if pathobj.as_posix()[0] == '$':
        return pathobj
    return pathobj.resolve()


@log_func_call(DEBUGLOW2, trace_only=True)
def get_expanded_pureposixpath(x: str | Path, addl_expand_vars: dict = {},
                               case_insensitive: bool = IS_WIN32):
    """
    Returns a `PurePosixPath` object equivalent to the given input path with
    all known variables of form `$var` or `${var}` expanded.
    See `get_equiv_pureposixpath()` for details on why this function is
    necessary.

    See `expandvars()` for details on how variables are expanded.

    The given path is merely expanded and returned as a `PurePosixPath`.
    User directories are not expanded, nor is the path resolved to an
    absolute one.

    Args:
        x (str  |  Path): input path
        addl_expand_vars (dict, optional): additional variables to override or
            supplement OS environment variables.  See `expandvars()` for
            details. Defaults to {}.

    Returns:
        PurePosixPath: path to `x` as a `PurePosixPath` with all known
            variables expanded, or None if `x` is None or empty
    """
    if x:
        return get_equiv_pureposixpath(expandvars(str(x), addl_expand_vars,
                                                  case_insensitive))


@log_func_call(DEBUGLOW2, trace_only=True)
def get_expanded_pathobj(x: str | Path, addl_expand_vars: dict = {},
                         case_insensitive: bool = IS_WIN32,
                         resolve: bool = True):
    """
    Returns an absolute `Path` object on the current system platform
    resolved from the given input path with user directories and all known
    variables of form `$var` or `${var}` expanded.

    See `expandvars()` for details on how variables are expanded.

    Args:
        x (str  |  Path): input path
        addl_expand_vars (dict, optional): additional variables to override or
            supplement OS environment variables.  See `expandvars()` for
            details. Defaults to {}.

    Returns:
        Path: fully-expanded absolute path to `x`, or None if `x` is None
            or empty
    """
    if x:
        ppp = get_expanded_pureposixpath(x, addl_expand_vars, case_insensitive)
        if resolve:
            return pureposixpath_to_resolved_pathobj(ppp)
        return pureposixpath_to_pathobj(ppp)


DLL_EXTS = ('.dll', '.so', '.dylib')


@log_func_call(DEBUGLOW2, trace_only=True)
def get_dll_ext_for_platform(platform: str = None):
    """
    Returns the preferred DLL file extension for the given
    `sys.platform` string

    Args:
        platform (str, optional): output of `sys.platform` to use.  Can be
            user-provided to allow overriding current platform for cross-target
            scripts.  If None, gets the current value of `sys.platform`.

    Returns:
        str: DLL file extension for platform
    """
    platform = platform or sys.platform

    if platform == 'win32':
        return '.dll'
    if platform == 'darwin':
        return '.dylib'
    return '.so'


@log_func_call(DEBUGLOW2, trace_only=True)
def replace_extension(p: Path | PurePath, ext: str):
    """
    Returns path `p` with the final suffix replaced by `ext`.

    Args:
        p (Path | PurePath): path to replace final suffix
        ext (str): extension to substitute for final suffix

    Returns:
        Path | PurePath: path with extension substituted
    """
    # return p.parent/(p.stem + ext)
    return p.with_suffix(ext)


@log_func_call(DEBUGLOW2, trace_only=True)
def test_alt_dll_paths(dll_path: Path, prefer_ext: str):
    """
    Test existence of the given DLL path with extensions other than the
    preferred extension.  Return the first found path.

    Args:
        dll_path (Path): path to replace extension and test
        prefer_ext (str): preferred extension (i.e. extension to skip testing)

    Returns:
        Path: found file that exists, otherwise None
    """
    prefer_ext = prefer_ext.lower()
    for ext in DLL_EXTS:
        if ext == prefer_ext:
            continue
        p = replace_extension(dll_path, ext)
        if p.exists():
            return p


@log_func_call(DEBUGLOW2, trace_only=True)
def expand_and_check_var(var_name: str, addl_expand_vars: dict = {},
                         case_insensitive: bool = IS_WIN32):
    "returns resolved, value"

    value = expandvars(f'${var_name}', addl_expand_vars, case_insensitive)
    resolved = is_key_resolved(value, var_name, case_insensitive)
    return resolved, value


@log_func_call(DEBUGLOW2, trace_only=True)
def expand_and_check_var_path(var_name: str, addl_expand_vars: dict = {},
                              case_insensitive: bool = IS_WIN32,
                              resolve_path: bool = True):
    "returns resolved, value"

    resolved, value = expand_and_check_var(var_name, addl_expand_vars,
                                           case_insensitive)
    if resolve_path:
        p = get_expanded_pathobj(value, addl_expand_vars, case_insensitive)

    else:
        ppp = get_expanded_pureposixpath(value, addl_expand_vars,
                                         case_insensitive)
        p = pureposixpath_to_pathobj(ppp)

    return resolved, p



from typing import TypeVar
from collections.abc import Container, Sequence

# from ..logging import DEBUGLOW2, log_func_call
from .constants import NODEFAULT, IS_WIN32

ContainerType = TypeVar('T', bound=Container)


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_value(x, case_insensitive: bool = IS_WIN32):
    if isinstance(x, str) and case_insensitive:
        return x.lower()
    return x


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_container(x: ContainerType, case_insensitive: bool = IS_WIN32
                       ) -> ContainerType:
    return type(x)(casesafe_value(z, case_insensitive) for z in x)


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_sequence_index(x: Sequence, value,
                            case_insensitive: bool = IS_WIN32):
    seq = casesafe_container(x, case_insensitive)
    return seq.index(casesafe_value(value, case_insensitive))


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_dict_key_map(d: dict, case_insensitive: bool = IS_WIN32):
    return {casesafe_value(x, case_insensitive): x for x in d}


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_dict_get(d: dict, key, default=NODEFAULT,
                      case_insensitive: bool = IS_WIN32):
    key = casesafe_value(key, case_insensitive)
    lookup = casesafe_dict_key_map(d, case_insensitive)
    return d[lookup[key]] if key in lookup else default


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_dict_set(d: dict, key, value, case_insensitive: bool = IS_WIN32):
    lookup = casesafe_dict_key_map(d, case_insensitive)
    key = casesafe_value(key, case_insensitive)
    key = lookup[key] if key in lookup else key
    d[key] = value


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_key_in_dict(d: dict, key, case_insensitive: bool = IS_WIN32):
    return casesafe_value_in_container(d, key, case_insensitive)


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_value_in_container(c: Container, key,
                                case_insensitive: bool = IS_WIN32):
    if isinstance(c, dict):
        c = tuple(casesafe_dict_key_map(c, case_insensitive).keys())
    return (casesafe_value(key, case_insensitive)
            in casesafe_container(c, case_insensitive))


# @log_func_call(DEBUGLOW2, trace_only=True)
def casesafe_is_equal(a, b, case_insensitive: bool = IS_WIN32):
    a = casesafe_value(a, case_insensitive)
    b = casesafe_value(b, case_insensitive)
    return a == b



from ..config.keys import LOCAL_CFG_KEY
from ..logging import log_exc, log_func_call, DEBUGLOW2, log_debug, log_info
from ..app import PyRandyOSApp
from . import GuiQtWrapper
from .styles import ThemeMap
from .window import GuiWindow
from .splash import GuiSplashScreen
from .loadstatus import load_status_step, splash_message
from .qrc import compile_qrc, import_qrc
from .qt import (
    QWidget, QApplication, qVersion, QObject, QEvent, QPixmap, Qt,
)

_GUI_APP_INST: 'GuiApp | None' = None


@log_func_call
def get_gui_app():
    global _GUI_APP_INST
    if not _GUI_APP_INST:
        raise RuntimeError("Qt application instance is not initialized.")
    return _GUI_APP_INST


class QtApp(QApplication):
    # only log user code, but can uncomment if you want to log all Qt events
    # @log_func_call(DEBUGLOW2)
    def notify(self, receiver: QObject, event: QEvent):
        try:
            # Add defensive checks to prevent the parameter mixup issue
            if isinstance(receiver, QEvent):
                receiver

            if not isinstance(receiver, QObject):
                raise RuntimeError("notify() called with invalid receiver "
                                   f"type: {type(receiver)}. Expected QObject")
            if not isinstance(event, QEvent):
                raise RuntimeError("notify() called with invalid event type: "
                                   f"{type(event)}. Expected QEvent")

            return super().notify(receiver, event)
        except BaseException as e:
            log_exc(e)
            return False


class GuiApp(GuiQtWrapper):
    INIT_GUI_IN_CONSTRUCTOR: bool = True

    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(self, app_args: list[str], *firstwin_args, **firstwin_kwargs):
        log_debug('initializing application')
        global _GUI_APP_INST
        if _GUI_APP_INST is not None:
            raise RuntimeError(
                "A Qt application instance already exists for this Python "
                "process. Only one instance is allowed per Python process."
            )
        _GUI_APP_INST = self
        self.gui_initialized = False

        super().__init__(None)
        self.qtobj: QApplication

        self.themes: ThemeMap = None
        self.windows: list[GuiWindow] = list()
        self.splash: GuiSplashScreen = None
        if self.INIT_GUI_IN_CONSTRUCTOR:
            self.init_gui(app_args, *firstwin_args, **firstwin_kwargs)

    @log_func_call
    def create_splash(self, *args, **kwargs) -> GuiSplashScreen:
        return GuiSplashScreen(QPixmap(), *args, **kwargs)

    @load_status_step("GUI initialized", show_step_done=True,
                      show_step_start=False)
    @log_func_call
    def init_gui(self, app_args: list[str], *firstwin_args, **firstwin_kwargs):
        log_debug('starting app main')

        # Ensure Qt resources are registered before any widgets are created
        compile_qrc()
        import_qrc()

        app = self.create_qt_inst(app_args)
        super().__init__(app)
        PyRandyOSApp.set('Qt_version', qVersion())

        splash = self.create_splash()
        if splash:
            splash.show()
            app.processEvents()
            self.splash = splash

        from .icons.iconfont import init_iconfonts
        init_iconfonts()

        self.init_themes()
        self.set_theme()
        # set_high_dpi_support(log=log)
        self.create_first_window(*firstwin_args, **firstwin_kwargs)
        self.gui_initialized = True

    def process_events(self):
        qtobj = self.qtobj
        if qtobj:
            qtobj.processEvents()

    @log_func_call
    def main(self, *args, **kwargs):
        if not self.gui_initialized:
            self.init_gui(*args, **kwargs)

        log_debug('starting Qt main loop')
        splash_message("Main loop starting")
        self.qtobj.exec_()

    @log_func_call
    @load_status_step("Initializing themes")
    def init_themes(self):
        self.themes = ThemeMap(self.qtobj)

    @log_func_call
    def set_theme(self, t: str = None):
        tnew = t
        t = t or PyRandyOSApp[f'{LOCAL_CFG_KEY}.theme']
        log_debug(f'setting theme to {t}')
        if tnew:
            PyRandyOSApp.set(f'{LOCAL_CFG_KEY}.theme', t)

        themes = self.themes
        themes.apply_theme()  # reset theme to default
        themes.apply_theme(t)

    @log_func_call
    def get_theme(self):
        return self.themes.get_current_theme()

    @log_func_call
    def create_qt_inst(self, app_args: list[str] = []):
        self.set_high_dpi_support()
        return QtApp(list(app_args))

    @log_func_call
    def create_first_window(self, *args, **kwargs):
        "create and show the first window after app launch"
        raise NotImplementedError('Abstract method not implemented')

    @log_func_call
    def close_splash(self, delegate: QWidget = None):
        if self.splash:
            qtsplash = self.splash.gui_view.qtobj
            qtsplash.finish(delegate)
            self.splash = None

    @log_func_call
    def set_high_dpi_support(self):
        # Attribute Qt::AA_EnableHighDpiScaling must be set before
        # QCoreApplication is created.
        try:
            QtApp.setAttribute(Qt.AA_EnableHighDpiScaling)
            QtApp.setAttribute(Qt.AA_UseHighDpiPixmaps)
        except Exception as e:
            # log_info("Failed to set high DPI support", exc_info=e)
            # on second thought, we don't need a full traceback
            log_info(f"Failed to set high DPI support: {e}")

    @log_func_call
    def get_dpi(self):
        return self.qtobj.primaryScreen().logicalDotsPerInch()

    @log_func_call
    def get_dpi_scale(self):
        return self.get_dpi() / 96.0



from typing import Generic

from ...logging import log_func_call
from ..qt import QLabel, QFrame, Qt, QGraphicsView, QStackedWidget
from .graphview import GraphViewWidget
from . import QtWidgetWrapper, QtWidgetType, GuiWidgetParentType


class GuiViewBaseWidget(Generic[QtWidgetType]):
    qtobj: QtWidgetType


class GuiViewBaseQtWidget(GuiViewBaseWidget[QtWidgetType],
                          QtWidgetWrapper[QtWidgetType],
                          Generic[QtWidgetType]):
    pass


class GuiViewBaseLabel(GuiViewBaseQtWidget[QLabel]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType, text: str = ""):
        super().__init__(gui_parent, text)

    @log_func_call
    def create_qtobj(self, text: str = ""):
        qtobj = QLabel(text, self.gui_parent.gui_view.qtobj)
        qtobj.setAlignment(Qt.AlignCenter)
        return qtobj


class GuiBaseGraphView(GuiViewBaseWidget[QGraphicsView], GraphViewWidget):
    pass


class GuiViewBaseFrame(GuiViewBaseQtWidget[QFrame]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType):
        super().__init__(gui_parent)

    @log_func_call
    def create_qtobj(self):
        return QFrame(self.gui_parent.gui_view.qtobj)


class GuiViewBaseStack(GuiViewBaseQtWidget[QStackedWidget]):
    @log_func_call
    def __init__(self, gui_parent: GuiWidgetParentType):
        super().__init__(gui_parent)

    @log_func_call
    def create_qtobj(self):
        return QStackedWidget(self.gui_parent.gui_view.qtobj)



from keyword import iskeyword


def legalize_iconname(iconname: str) -> str:
    """
    Convert an icon name to a legal Python identifier.
    """
    newname = iconname.replace('-', '_')
    if iskeyword(newname):
        newname += '_'

    if newname[0].isdigit():
        newname = f'_{newname}'

    if newname == "l":
        newname = "L"

    return newname



# adapted from code in Python stdlib posixpath.py
from typing import Any
from collections.abc import Callable
from re import compile, ASCII
from functools import partial
from os import environ

from ..logging import log_func_call, DEBUGLOW2
from .cfgdict import config_dict_get
from .constants import IS_WIN32
from .casesafe import casesafe_is_equal, casesafe_value, casesafe_key_in_dict

_VARPROG = compile(r'\$(\w+|\{[^}]*\})', ASCII)
_STARTBRAK = '{'
_ENDBRAK = '}'
_NOTFOUND = object()


@log_func_call(DEBUGLOW2, trace_only=True)
def expandvars_base(x: str, callback: Callable) -> str:
    """
    Invokes `callback()` when string `x` contains a variable of the form
    "$variable" or "${variable}" and returns the resultant string after
    callbacks.

    In the context of `expandvars()`, the callable substitutes the variable
    with a value, but `expandvars_base()` allows the user to customize this
    behavior.  Example alternative use cases could be to simply collect a list
    of all the variables in the string and perform no substitutions or to do
    some sort of conditional substitution.

    The user-provided `callback` should have the signature
    `callback(data: dict)`.  The argument `data` is a dictionary with the
    following keys/values:

        * `x`: current working version of the full modified string
        * `i`: index of the `$` character in `x` for the current variable upon
            which the callback is to operate
        * `j`: index of the first character in `x` following the current
            variable upon which the callback is to operate
        * `name`: the name of the variable

    The callback should return True if `expandvars_base()` should continue
    searching for more variables in `x` or False to end the search early and
    return `x` in its current state.

    See the code for `expandvars()` or `get_unresolved_keys()` for complete
    usage examples.

    Args:
        x (str): input string to search for variables to expand
        callback (Callable): user-provided callback with signature
            `callback(data: dict) -> bool`

    Returns:
        str: value of `x` after performing callbacks
    """
    search = _VARPROG.search
    data = {'x': x, 'i': 0}
    ok_to_continue = True
    while ok_to_continue:
        m = search(data['x'], data['i'])
        if not m:
            break

        data['i'], data['j'] = m.span(0)
        name = m.group(1)
        if name.startswith(_STARTBRAK) and name.endswith(_ENDBRAK):
            name = name[1:-1]

        data['name'] = name
        ok_to_continue = callback(data)
        if ok_to_continue is None:
            ok_to_continue = True

    return data['x']


@log_func_call(DEBUGLOW2, trace_only=True)
def substitute_callback(key: str, subst: str, case_insensitive: bool,
                        data: dict):
    if casesafe_is_equal(data['name'], key, case_insensitive):
        x = data['x']
        tail = x[data['j']:]
        x = x[:data['i']] + str(subst)
        data['i'] = len(x)
        data['x'] = x + tail
        return False
    return True


@log_func_call(DEBUGLOW2, trace_only=True)
def substitute_key(value: str, key: str, subst: str,
                   case_insensitive: bool = IS_WIN32):
    return expandvars_base(value, partial(substitute_callback, key, subst,
                                          case_insensitive))


@log_func_call(DEBUGLOW2, trace_only=True)
def unresolved_keys_callback(keylist: list, data: dict):
    keylist.append(data['name'])
    data['i'] = data['j']
    return True


@log_func_call(DEBUGLOW2, trace_only=True)
def get_unresolved_keys(x: str):
    keys: list[str] = list()
    expandvars_base(x, partial(unresolved_keys_callback, keys))
    return keys


@log_func_call(DEBUGLOW2, trace_only=True)
def expandvars_callback(addl_expand_vars: dict, case_insensitive: bool,
                        data: dict):
    value = _NOTFOUND
    name: str = casesafe_value(data['name'], case_insensitive)
    if '.' in name or casesafe_key_in_dict(addl_expand_vars, name,
                                           case_insensitive):
        value = config_dict_get(addl_expand_vars, name, _NOTFOUND,
                                case_insensitive)

    if value is _NOTFOUND:
        try:
            value = environ[name]
        except KeyError:
            pass

    if value is _NOTFOUND:
        value = None

    if value is None:
        data['i'] = data['j']

    else:
        substitute_callback(name, value, case_insensitive, data)


@log_func_call(DEBUGLOW2, trace_only=True)
def expandvars(x: str, addl_expand_vars: dict[str, Any] = {},
               case_insensitive: bool = IS_WIN32) -> str:
    """
    Expand variables of form $var and ${var}.  Unknown variables
    are left unchanged.  Searches for `var` in `addl_expand_vars` first, then
    checks `os.environ`.  This way, env vars can be overridden locally or
    additional context vars can be used.

    Args:
        x (str): string to search for variables in which to substitute
        addl_expand_vars (dict, optional): additional variables to override or
            supplement the current OS environment variables. Defaults to {}.

    Returns:
        str: string with all known variables substituted
    """

    if '$' not in x:
        return x
    return expandvars_base(x, partial(expandvars_callback, addl_expand_vars,
                                      case_insensitive))


@log_func_call(DEBUGLOW2, trace_only=True)
def is_key_resolved(x: str, key: str, case_insensitive: bool = IS_WIN32):
    return not casesafe_key_in_dict(get_unresolved_keys(x), key,
                                    case_insensitive)



from json import loads as jloads, JSONDecodeError, dumps as jdumps
from pathlib import Path
from types import NoneType
from re import sub

from ..logging import log_func_call

JSONTYPES = (str, float, int, bool, list, dict, NoneType)
JsonDataType = str | float | int | bool | list | dict | NoneType


@log_func_call
def load_jsonc(fn: str | Path) -> dict | list:
    """
    Open a JSON file and return its parsed contents as a dict or list.

    Args:
        fn (str | Path): file path to open

    Raises:
        JSONDecodeError: provides details on what file it failed to parse and
            details about the error.

    Returns:
        dict | list: parsed contents of the file `fn`
    """
    try:
        return parse_jsonc(Path(fn).read_text())
    except JSONDecodeError as e:
        raise JSONDecodeError(f'error reading {fn}: {e.msg}', e.doc, e.pos)


@log_func_call
def parse_jsonc(jsonstr: str) -> dict | list:
    """
    return the parsed contents of given string as a dict or list.

    Args:
        jsonstr (str): JSON-formatted string to parse

    Raises:
        JSONDecodeError: provides details on why it failed to parse

    Returns:
        dict | list: parsed contents of the string `jsonstr`
    """
    buf = ""
    for line in jsonstr.splitlines():
        curline = ''
        inquote = False
        escape = False
        lastc = ''
        for c in line:
            if escape:
                escape = False
            elif c == '"':
                inquote = not inquote
            elif inquote:
                if c == '\\':
                    escape = True
            elif c == '/' and lastc == '/':
                lastc = ''
                break
            curline += lastc
            lastc = c
        curline += lastc
        buf += curline

    # Remove trailing commas in objects and arrays
    buf = sub(r',([ \t\r\n]*[}}\]])', r'\1', buf)

    return jloads(buf)


@log_func_call
def save_json(fn: str | Path, data: dict | list):
    """
    Save `data` to the file `fn` in JSON format.

    The values inside of `data` must all be JSON-compatible types.
    The constant `JSON_TYPES` lists all valid pure JSON types.  While the
    contents of `data` do not need to be exclusively these types, they must be
    easily coercible to one of these types.  The validation of types is done
    by the Python standard library JSON implementation and no additional
    validation is performed by this function.

    Args:
        fn (str | Path): file path to save
        data (dict | list): structured data to convert to JSON.
    """
    fn.write_text(jdumps(data, indent=2))


@log_func_call
def jsonify(data) -> JsonDataType:
    if isinstance(data, dict):
        return {k: jsonify(v) for k, v in data.items()}

    elif isinstance(data, list):
        return [jsonify(item) for item in data]

    elif isinstance(data, (str, float, int, bool)) or data is None:
        return data

    elif isinstance(data, Path):
        return data.as_posix()

    else:
        raise TypeError(f"Type is not JSON serializable: {type(data)}")



from logging import Logger
from pathlib import Path
from copy import deepcopy

from ..logging import (
    DEBUGLOW2, log_func_call, set_trace_logging, set_func_call_logging,
    set_global_logger, get_global_logger
)
from ..utils.classproperty import classproperty
from ..utils.paths import (
    get_expanded_pathobj, get_expanded_pureposixpath, pureposixpath_to_pathobj
)
from ..utils.json import load_jsonc
from ..utils.cfgdict import (
    config_dict_get, config_dict_set, config_dict_update,
)
from ..utils.constants import NODEFAULT, IS_WIN32
from ..utils.system import build_cmd_arg_list
from ..utils.casesafe import casesafe_is_equal, casesafe_value_in_container
from ..utils.stack import set_show_traceback_locals

from .defaults import get_defaults
from .keys import (
    BASE_LOG_DIR_KEY, BASE_PATH_KEY, LOG_TIMESTAMP_KEY,
    APPEND_LOG_KEY, CLI_LOG_LEVEL_KEY, get_path_keys, BASE_LOG_PATH_KEY,
    FILE_LOG_LEVEL_KEY, ABS_BASE_PATH_KEY, LOG_TRACE_ENABLED_KEY,
    SHOW_TRACEBACK_LOCALS_KEY, LOG_FUNC_CALL_ENABLED_KEY,
)
from .expandutils import expand_key_recursively

_GLOBAL_CFG: dict = None


class AppConfigType(type):
    # this is really just here for VSCode syntax highlighting because the real
    # class won't necessarily have access to this as a metaclass.
    # Don't overthink this.  Just make sure it's set in AppConfig and
    # its subclasses.
    CONFIG_CASE_INSENSITIVE: bool

    @log_func_call(DEBUGLOW2, trace_only=True)
    def get_case(cls, case_insensitive: bool = None):
        return (cls.CONFIG_CASE_INSENSITIVE if case_insensitive is None
                else case_insensitive)

    @log_func_call(DEBUGLOW2, trace_only=True)
    def __contains__(cls, item):
        return casesafe_value_in_container(_GLOBAL_CFG, item, cls.get_case())


class AppConfig(metaclass=AppConfigType):
    CONFIG_CASE_INSENSITIVE: bool = IS_WIN32

    @log_func_call
    def __init__(self):
        # Class is intended as a singleton
        if type(self) is AppConfig:  # pragma: no cover
            raise NotImplementedError('Singleton class should not be '
                                      'instantiated')

    @classproperty
    @log_func_call(DEBUGLOW2, trace_only=True)
    def log(cls):
        return get_global_logger()

    @classmethod
    @log_func_call
    def set_logger(cls, log: Logger):
        set_global_logger(log)

    @classmethod
    @log_func_call
    def set(cls, key: str, value, case_insensitive: bool = None):
        config_dict_set(_GLOBAL_CFG, key, value,
                        cls.get_case(case_insensitive))

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def __class_getitem__(cls, key: str):
        return config_dict_get(_GLOBAL_CFG, key,
                               case_insensitive=cls.get_case())

    @classmethod
    @log_func_call
    def set_global_config(cls, config: dict):
        log = get_global_logger()
        log.debug('setting global config')
        global _GLOBAL_CFG
        _GLOBAL_CFG = config

    @classmethod
    @log_func_call(DEBUGLOW2)
    def get_global_config(cls):
        return _GLOBAL_CFG

    @classproperty
    @log_func_call(DEBUGLOW2)
    def global_config(cls):
        return cls.get_global_config()

    @global_config.setter
    @log_func_call
    def global_config(cls, config: dict):
        cls.set_global_config(config)

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def get(cls, key: str, default=NODEFAULT,
            case_insensitive: bool = None):
        """
        If the global app config has been set, returns the value for the
        given key if present, else returns the default if provided.
        When no global app config is set, returns None.

        If the key contains '.', it assumes that the key is a dot-delimited
        list of keys of nested dicts or lists, and will recursively drill down
        through the config dict to return the ultimate value.

        Args:
            key (str): key to get from dict or dot-delimited list of keys
            default (Any, optional): Value to return if key not found.
                Defaults to NODEFAULT, which raises a KeyError if key is
                not found.

        Raises:
            KeyError: key not found in `config` or its nested children

        Returns:
            Any: value for key or default value (if provided) or None if the
                global app config is not set.
        """
        if _GLOBAL_CFG:
            return config_dict_get(_GLOBAL_CFG, key, default,
                                   cls.get_case(case_insensitive))

    @classmethod
    @log_func_call
    def update(cls, data: dict, case_insensitive: bool = None):
        config_dict_update(_GLOBAL_CFG, data, cls.get_case(case_insensitive))

    @classmethod
    @log_func_call
    def build_cmd_args_from_config(cls, key: str):
        return build_cmd_arg_list(cls[key])

    @classmethod
    @log_func_call
    def process_config(cls, skip_expansion: str | list[str] = 'skip_expand',
                       config: dict = None, app_path_keys: tuple[str] = (),
                       case_insensitive: bool = None):
        config = config or _GLOBAL_CFG
        case = cls.get_case(case_insensitive)
        skip_expansion = skip_expansion or ()
        if isinstance(skip_expansion, str):
            skip_expansion = (skip_expansion,)

        for k in config.keys():
            # print(k)
            expand_key_recursively(config, k, skip_expansion,
                                   case_insensitive=case)

        base_path = get_expanded_pathobj(config[BASE_PATH_KEY], config, case)
        config_dict_set(config, BASE_PATH_KEY, base_path, case)
        config_dict_set(config, ABS_BASE_PATH_KEY, base_path.absolute(), case)
        for k in get_path_keys(app_path_keys):
            if casesafe_is_equal(k, BASE_PATH_KEY):
                continue

            v = config_dict_get(config, k, None, case)
            if v:
                config_dict_set(config, k, cls.handle_path(v, base_path))

        set_trace_logging(config_dict_get(config, LOG_TRACE_ENABLED_KEY,
                                          case_insensitive=case))
        set_func_call_logging(config_dict_get(config,
                                              LOG_FUNC_CALL_ENABLED_KEY,
                                              case_insensitive=case))
        set_show_traceback_locals(config_dict_get(config,
                                                  SHOW_TRACEBACK_LOCALS_KEY,
                                                  case_insensitive=case))
        return config

    @classmethod
    @log_func_call
    def get_base_logpath(cls):
        return config_dict_get(_GLOBAL_CFG, BASE_LOG_PATH_KEY,
                               case_insensitive=cls.get_case())

    @classmethod
    @log_func_call
    def init_parse_config(cls, indata: dict | str | Path = None,
                          overrides: dict = None, defaults: dict = None,
                          app_global_defaults: dict = {},
                          app_local_defaults: dict = {},
                          case_insensitive: bool = None):
        defaults = defaults or get_defaults(cls, app_global_defaults,
                                            app_local_defaults)

        config = deepcopy(defaults)
        if indata:
            config_dict_update(config, indata if isinstance(indata, dict)
                               else load_jsonc(indata), case_insensitive)

        if overrides:
            config_dict_update(config, overrides, case_insensitive)

        global _GLOBAL_CFG
        _GLOBAL_CFG = config

    @classmethod
    @log_func_call
    def expand_log_config(cls, skip_expansion: str | list[str] = None):
        # An abbreviated function to get the log path quickly before further
        # expansions so further error messages can be logged.
        # All of these keys are hardcoded, so we don't need to worry about case
        config = _GLOBAL_CFG
        expand_key_recursively(config, BASE_PATH_KEY, skip_expansion)
        expand_key_recursively(config, BASE_LOG_DIR_KEY, skip_expansion)
        expand_key_recursively(config, LOG_TIMESTAMP_KEY, skip_expansion)
        expand_key_recursively(config, APPEND_LOG_KEY, skip_expansion)
        expand_key_recursively(config, CLI_LOG_LEVEL_KEY, skip_expansion)
        expand_key_recursively(config, FILE_LOG_LEVEL_KEY, skip_expansion)
        expand_key_recursively(config, LOG_TRACE_ENABLED_KEY, skip_expansion)
        expand_key_recursively(config, LOG_FUNC_CALL_ENABLED_KEY,
                               skip_expansion)
        expand_key_recursively(config, SHOW_TRACEBACK_LOCALS_KEY,
                               skip_expansion)
        base_path = get_expanded_pathobj(config[BASE_PATH_KEY], config)
        logdir = config_dict_get(config, BASE_LOG_DIR_KEY)
        timestamp_name = config_dict_get(config, LOG_TIMESTAMP_KEY)
        append_log = config_dict_get(config, APPEND_LOG_KEY)
        cli_log_level = config_dict_get(config, CLI_LOG_LEVEL_KEY)
        file_log_level = config_dict_get(config, FILE_LOG_LEVEL_KEY)
        log_trace_enabled = config_dict_get(config, LOG_TRACE_ENABLED_KEY)
        log_func_call_enabled = config_dict_get(config,
                                                LOG_FUNC_CALL_ENABLED_KEY)
        tb_locals_enabled = config_dict_get(config, SHOW_TRACEBACK_LOCALS_KEY)
        return (cls.handle_path(logdir, base_path), timestamp_name,
                append_log, cli_log_level, file_log_level, log_trace_enabled,
                tb_locals_enabled, log_func_call_enabled)

    @classmethod
    @log_func_call
    def handle_path(cls, x: str, base_path: Path,
                    case_insensitive: bool = None) -> Path:
        newppp = get_expanded_pureposixpath(x, _GLOBAL_CFG, case_insensitive)
        newpath = pureposixpath_to_pathobj(newppp)

        # any relative paths are assumed to be wrt base path for now
        newpathstr = str(newpath)
        if not newpath.is_absolute() and newpathstr[0] != '$':
            if x[:2] in ('./', '.\\'):
                newpath = get_expanded_pathobj(newpath)
            else:
                newpath = base_path/newpath

        return newpath

    @staticmethod
    @log_func_call
    def get_local_config(base: dict = None, case_insensitive: bool = None):
        from .local import get_local_config as _getlocal
        return _getlocal(base, case_insensitive)

    @staticmethod
    @log_func_call
    def save_local_config(app_path_keys: tuple = (),
                          case_insensitive: bool = None):
        from .local import save_local_config as _savelocal
        _savelocal(app_path_keys, case_insensitive)



"""
QT MODERN

MIT License

Copyright (c) 2017 Gerard Marull-Paretas <gerardmarull@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
from re import sub, M as re_M

try:
    from qdarkstyle import load_stylesheet as load_dark_stylesheet  # type: ignore  # noqa: E501
    HAS_QDARKSTYLE = True
except ImportError:
    def load_dark_stylesheet():
        pass

    HAS_QDARKSTYLE = False

from ....logging import log_func_call, log_warning
from ...qt import QPalette, QColor
if TYPE_CHECKING:
    from ...gui_app import QtApp

from .vibedark import vibedark


@log_func_call
def qdarkstyle(app: QtApp):
    """
    Apply dark theme to the Qt application instance.

    Args:
        app (QtApp): QApplication instance.
    """

    if not HAS_QDARKSTYLE:
        log_warning("QDarkStyle is not installed. Downmoding to VibeDark")
        return vibedark(app)

    dark_palette = QPalette()

    # base
    dark_palette.setColor(QPalette.WindowText, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.Button, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.Light, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.Midlight, QColor(90, 90, 90))
    dark_palette.setColor(QPalette.Dark, QColor(35, 35, 35))
    dark_palette.setColor(QPalette.Text, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.BrightText, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.ButtonText, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.Base, QColor(42, 42, 42))
    dark_palette.setColor(QPalette.Window, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    dark_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.HighlightedText, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.Link, QColor(56, 252, 196))
    dark_palette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
    dark_palette.setColor(QPalette.ToolTipBase, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ToolTipText, QColor(180, 180, 180))
    dark_palette.setColor(QPalette.LinkVisited, QColor(80, 80, 80))

    # disabled
    dark_palette.setColor(QPalette.Disabled, QPalette.WindowText, QColor(127, 127, 127))  # noqa: E501
    dark_palette.setColor(QPalette.Disabled, QPalette.Text, QColor(127, 127, 127))  # noqa: E501
    dark_palette.setColor(QPalette.Disabled, QPalette.ButtonText, QColor(127, 127, 127))  # noqa: E501
    dark_palette.setColor(QPalette.Disabled, QPalette.Highlight, QColor(80, 80, 80))  # noqa: E501
    dark_palette.setColor(
        QPalette.Disabled, QPalette.HighlightedText, QColor(127, 127, 127)
    )

    app.style().unpolish(app)
    app.setPalette(dark_palette)
    app.setStyle("Fusion")

    # fix for QComboBox indentation issue
    # see: https://github.com/ColinDuquesnoy/QDarkStyleSheet/issues/308#issuecomment-1661775924  # noqa: E501
    qss = load_dark_stylesheet(qt_api='pyside2')
    qss = sub(r'QComboBox::indicator([^}]|\n)+}', '', qss, re_M)
    app.setStyleSheet(qss)



import sys
from pathlib import Path
from tempfile import gettempdir

from .logging import (
    log_info, Logger, log_func_call, DEBUGLOW2, set_trace_logging,
    set_func_call_logging,
)
from .config import AppConfig
from .config.keys import (
    BASE_LOG_PATH_KEY, APP_NAME_KEY, get_log_dir_keys, APP_PKG_DIR_KEY,
    APP_ASSETS_DIR_KEY, APP_PKG_VERSION_KEY, TMP_DIR_KEY, LOCAL_CFG_KEY
)
from .config.local import process_local_config

from .utils.constants import DEFAULT_GROUP, DEFAULT_DIR_MODE
from .utils.log import setup_logging, create_log_file
from .utils.main import MainContext
from .utils.system import mkdir_chgrp
from .utils.stack import (
    top_package_dir_path, top_module_and_name, set_show_traceback_locals
)


class PyRandyOSApp(AppConfig):
    # abstract class attributes
    APP_NAME: str
    APP_LOG_PREFIX: str

    # class attributes with fallback defaults
    APP_PATH_KEYS: tuple[str] = ()
    APP_LOG_DIR_KEYS: tuple[str] = ()
    APP_GLOBAL_DEFAULTS = {}
    APP_LOCAL_DEFAULTS = {}
    APP_ASSETS_DIR: str | Path = None
    "path to be resolved when config is processed; may use variable expansion"

    use_local_config: bool = None

    @classmethod
    @log_func_call
    def main(cls, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    @log_func_call
    def init_main(cls, config: dict | str | Path = None,
                  setup_log: bool = False, logfile: Path = None,
                  logger: Logger = None, **kwargs):
        "returns True if a local config is present and loaded, else False"
        # log = logger or get_logger()
        cls.set_logger(logger)
        cls.init_parse_config(config, kwargs)

        # setup logging first if necessary:
        if setup_log:
            (
                logdir,
                ts_name,
                append,
                cli_log_level,
                file_log_level,
                log_trace_enabled,
                tb_locals_enabled,
                log_func_call_enabled,
            ) = cls.expand_log_config()
            logfile = logfile or create_log_file(logdir, ts_name, append,
                                                 cls.APP_LOG_PREFIX)

            set_trace_logging(log_trace_enabled)
            set_func_call_logging(log_func_call_enabled)
            set_show_traceback_locals(tb_locals_enabled)
            setup_logging(logfile, cli_log_level, file_log_level)

        # start logging and process the rest of the configuration data
        cls.set(BASE_LOG_PATH_KEY, logfile)
        cls.set(APP_PKG_VERSION_KEY, cls.get_package_version())
        pkgdir = cls.get_package_dir()
        cls.set(APP_PKG_DIR_KEY, pkgdir)
        assets_dir = cls.get_assets_dir()
        if assets_dir:
            cls.set(APP_ASSETS_DIR_KEY, assets_dir)

        appname = cls.APP_NAME
        cls.set(APP_NAME_KEY, appname)
        log_info(f"Starting {appname}")
        cls.process_config()

        use_local_config = process_local_config(app_path_keys=cls.APP_PATH_KEYS)  # noqa: E501
        cls.use_local_config = use_local_config
        return use_local_config

    @classmethod
    @log_func_call
    def process_config(cls, skip_expansion: str | list[str] = 'skip_expand',
                       config: dict = None):
        return super().process_config(skip_expansion, config,
                                      app_path_keys=cls.APP_PATH_KEYS)

    @classmethod
    @log_func_call
    def init_parse_config(cls, indata: dict | str | Path = None,
                          overrides: dict = None, defaults: dict = None):
        global_defaults = dict()
        global_defaults.update(cls.APP_GLOBAL_DEFAULTS)
        return super().init_parse_config(indata, overrides, defaults,
                                         global_defaults,
                                         cls.APP_LOCAL_DEFAULTS)

    @classmethod
    @log_func_call
    def run_cmdline(cls, args: list[str] = None):
        if args is None:
            args = sys.argv[1:]

        with MainContext(cls.APP_NAME):
            sys.exit(cls.main(*cls.preprocess_args(args)))

    @classmethod
    @log_func_call
    def preprocess_args(cls, args: list[str]):
        if len(args) > 0:
            raise ValueError('too many command line arguments')

        return args

    @classmethod
    @log_func_call
    def create_log_dirs(cls, group: str = DEFAULT_GROUP,
                        mode: int = DEFAULT_DIR_MODE):
        for key in get_log_dir_keys(cls.APP_LOG_DIR_KEYS):
            p: Path = cls.get(key)
            mkdir_chgrp(p, group, mode)

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def mkdir_temp(cls, name: str = None, group: str = DEFAULT_GROUP,
                   mode: int = DEFAULT_DIR_MODE):
        tmp_dir: Path | None = cls.get(TMP_DIR_KEY)
        if not tmp_dir:
            try:
                appname = cls.APP_NAME
            except AttributeError:
                appname = 'PyRandyOSApp'

            tmp_dir = Path(gettempdir())/appname
            if cls.global_config:
                cls.set(TMP_DIR_KEY, tmp_dir)

        if name:
            tmp_dir = tmp_dir/name

        mkdir_chgrp(tmp_dir, group, mode)
        return tmp_dir

    @classmethod
    @log_func_call
    def get_package_version(cls):
        return getattr(top_module_and_name(cls)[0], '__version__')

    @classmethod
    @log_func_call
    def get_package_dir(cls):
        return top_package_dir_path(cls)

    @classmethod
    @log_func_call
    def get_assets_dir(cls):
        assetsdir = cls.get(APP_ASSETS_DIR_KEY, cls.APP_ASSETS_DIR)
        if assetsdir:
            return Path(assetsdir)

    @classmethod
    @log_func_call(DEBUGLOW2)
    def get_default_win_size(cls):
        return (cls[f'{LOCAL_CFG_KEY}.default_width'],
                cls[f'{LOCAL_CFG_KEY}.default_height'])

    @classmethod
    @log_func_call(DEBUGLOW2)
    def get_local_config(cls, case_insensitive: bool = None):
        return super().get_local_config(case_insensitive=case_insensitive)

    @classmethod
    @log_func_call
    def save_local_config(cls, case_insensitive: bool = None):
        return super().save_local_config(cls.APP_PATH_KEYS, case_insensitive)
